#pragma once

#if defined(__cplusplus)
extern "C" {
#endif

#include <lektor/internal/os.h>
#include <lektor/common.h>
#include <stddef.h>

typedef enum {
    URI_NULL = 0,
    URI_ID,
    URI_PLAYLIST,
    URI_TYPE,
    URI_AUTHOR,
    URI_CATEGORY,
    URI_LANGUAGE,
    URI_QUERY,
} LKT_URI_TYPE;

typedef enum {
    URI_VALUE_TYPE_NULL = 0,
    URI_VALUE_TYPE_STRING,
    URI_VALUE_TYPE_INTEGER,
} LKT_URI_VALUE_TYPE;

struct lkt_uri;       /* Hide the implementation of the lkt_uri structure */
struct kara_metadata; /* Forward declaration of the kara metadata from <lektor/mkv.h> */

/* Allocated a new empty URI. Never returns NULL */
struct lkt_uri *lkt_uri_new(void);

/* Init the URI with a string. Returns true if the creation was successfull,
 * false otherwise. */
bool lkt_uri_from_single(struct lkt_uri *ret, char *const str);

/* Init the URI with a virutal string which is the concatenation of all the
 * strings in the null terminated list where each element is separated by a
 * space.  Returns true if the creation was successfull, false otherwise. */
bool lkt_uri_from_list(struct lkt_uri *ret, char **list);

bool lkt_uri_from_s_id(struct lkt_uri *ret, int id);
bool lkt_uri_from_u_id(struct lkt_uri *ret, unsigned int id);
bool lkt_uri_from_l_id(struct lkt_uri *ret, long id);
bool lkt_uri_from_ul_id(struct lkt_uri *ret, unsigned long id);

// clang-format off
#define lkt_uri_from_id(ret, from)                      \
    _Generic((from), int           : lkt_uri_from_s_id, \
                     unsigned int  : lkt_uri_from_u_id, \
                     long          : lkt_uri_from_l_id, \
                     unsigned long : lkt_uri_from_u_id) \
    (ret, from)
// clang-format on

/* Free an URI. Any URI allocated with `lkt_uri_new` or `lkt_uri_copy` and any
 * other where one of the *from* function was called upon must be freed by this
 * function. The free function never fails. */
void lkt_uri_free(struct lkt_uri *ret);

LKT_URI_TYPE lkt_uri_get_type(const struct lkt_uri *);
const char *lkt_uri_get_value_as_str(const struct lkt_uri *);
int lkt_uri_get_value_as_int(const struct lkt_uri *);
LKT_URI_VALUE_TYPE lkt_uri_get_value_type(const struct lkt_uri *);
const char *lkt_uri_get_column_name(const struct lkt_uri *);

/* Copy an URI, the returned memory is allocated and must be freed with
 * lkt_uri_free. Never return NULL. */
struct lkt_uri *lkt_uri_copy(const struct lkt_uri *);

// clang-format off
#define lkt_uri_from(ret, from)                     \
    _Generic((from), char *  : lkt_uri_from_single, \
                     char ** : lkt_uri_from_list)   \
    (ret, from)
// clang-format on

/* Make an URL to download from kurisu.  base_url is the prefix for the api,
 * e.g. https://kurisu.iiens.net/api_kara.php */
bool lkt_uri_to_url(const struct lkt_uri *, const char *base_url, char *ret, size_t len);

/* Turn the URI into a human readable string. */
bool lkt_uri_to_str(const struct lkt_uri *, char *ret, size_t len);

/* Returns `true` if the uri matches the metadata. Returns false otherwise. */
bool lkt_uri_match_kara_metadata(const struct lkt_uri *, const struct kara_metadata *kara_mdt,
                                 lkt_db *, int kara_id);

#if defined(__cplusplus)
}
#endif
