use crate::{cards::*, constants, utils, widgets};
use amadeus_lib::{
    actions,
    deamon::{self, CommandDeamon, Deamon, StatusDeamon},
};
use amadeus_macro::*;
use eframe::{egui, App};
use lkt_lib::{
    Kara, LektorCurrentKaraInnerResponse, LektorPlaybackStatusResponse, LektorQuery,
    LektorResponse, LektorState, Playlist,
};
use log::*;
use std::{
    sync::mpsc::{Receiver, SendError, Sender},
    time,
};

type CommandDeamonData = (
    (Sender<LektorQuery>, Receiver<LektorResponse>),
    deamon::CommandDeamon,
);

type StatusDeamonData = (
    (Receiver<deamon::StatusDeamonMessageType>,),
    deamon::StatusDeamon,
);

pub struct Amadeus<'a> {
    config: utils::AmadeusConfig,
    has_config_changed: bool,
    need_about_window: bool,
    need_settings_window: bool,
    actions: Vec<actions::Action>,

    last_render_instant: time::SystemTime,
    begin_render_instant: time::SystemTime,

    deamon: Option<CommandDeamonData>,
    status_deamon: Option<StatusDeamonData>,

    lektord_status: Option<LektorPlaybackStatusResponse>,
    lektord_current: Option<LektorCurrentKaraInnerResponse>,

    lektord_queue: KaraCardCollection<'a>,
    lektord_historic: KaraCardCollection<'a>,

    lektord_search_results: KaraCardCollection<'a>,
    lektord_search_query: String,
    lektord_updated_query: bool,

    playlist_store: widgets::PlaylistsStore,

    lektord_state: LektorState,

    amadeus_logo_texture: Option<egui::TextureHandle>,
}

fn handle_sender_error(e: Result<(), SendError<LektorQuery>>) {
    if let Err(e) = e {
        let query = e.0;
        error!("error while sending commands {query:?} to command deamon")
    }
}

impl Default for Amadeus<'_> {
    fn default() -> Self {
        Self {
            last_render_instant: time::SystemTime::now(),
            begin_render_instant: time::SystemTime::UNIX_EPOCH,

            lektord_queue: KaraCardCollection::new("Queue".to_owned())
                .add_action(actions::Action::PlayFromKara)
                .add_action(actions::Action::DeleteKaraFromQueue)
                .add_action(actions::Action::AddKaraToPlaylist),
            lektord_historic: KaraCardCollection::new("Historic".to_owned())
                .add_action(actions::Action::AddKaraToQueue)
                .add_action(actions::Action::InsertKaraInQueue)
                .add_action(actions::Action::DeleteKaraFromQueue)
                .add_action(actions::Action::AddKaraToPlaylist),
            lektord_search_results: KaraCardCollection::new("Search results".to_owned())
                .add_action(actions::Action::AddKaraToQueue)
                .add_action(actions::Action::InsertKaraInQueue)
                .add_action(actions::Action::DeleteKaraFromQueue)
                .add_action(actions::Action::AddKaraToPlaylist),
            lektord_search_query: String::new(),

            actions: Vec::with_capacity(10),
            playlist_store: Default::default(),
            config: Default::default(),
            has_config_changed: Default::default(),
            need_about_window: Default::default(),
            need_settings_window: Default::default(),
            deamon: Default::default(),
            status_deamon: Default::default(),
            lektord_updated_query: Default::default(),
            lektord_state: Default::default(),
            amadeus_logo_texture: Default::default(),

            lektord_status: None,
            lektord_current: None,
        }
    }
}

impl Amadeus<'_> {
    pub fn create(cc: &eframe::CreationContext<'_>) -> Box<Self> {
        let mut ret = Amadeus::default();
        let ctx = &cc.egui_ctx;
        if let Some(storage) = cc.storage {
            ret.config = eframe::get_value(storage, "amadeus-rs-config").unwrap_or_default();
        }

        ctx.set_fonts(utils::font::get_font_definitions());
        ret.load_amadeus_logo(ctx);

        let mut style = (*ctx.style()).clone();
        style.text_styles = utils::font::get_font_styles();
        ctx.set_style(style);

        ret.apply_settings();

        Box::new(ret)
    }

    fn load_amadeus_logo(&mut self, ctx: &egui::Context) {
        let (logo_buffer, [size_x, size_y]) = utils::get_icon_as_dynamic_image();
        let pixels = logo_buffer.as_flat_samples();
        let logo_texture = egui::ColorImage::from_rgba_unmultiplied(
            [size_x as usize, size_y as usize],
            pixels.as_slice(),
        );
        self.amadeus_logo_texture =
            Some(ctx.load_texture("amadeus-logo", logo_texture, egui::TextureOptions::LINEAR));
    }

    fn collect_fulfilled_actions(&mut self) -> Vec<(u64, actions::Action)> {
        let mut ret = self.lektord_queue.fulfilled_actions();
        ret.extend(self.lektord_historic.fulfilled_actions());
        ret.extend(self.lektord_search_results.fulfilled_actions());
        ret.extend(self.playlist_store.fulfilled_actions());
        ret
    }

    fn render_side_panel_main_view_button(
        &mut self,
        ui: &mut egui::Ui,
        title: &str,
        view: utils::AmadeusMainView,
    ) {
        ui.add_space(constants::PADDING);
        let selected = view == self.config.main_panel_view;
        if selected {
            ui.colored_label(constants::get_text_color(self.config.dark_mode), title);
        } else {
            let the_btn = egui::Button::new(title).frame(false);
            if ui.add(the_btn).clicked() {
                self.config.main_panel_view = view;
            }
        }
    }

    fn render_side_panel(&mut self, ctx: &egui::Context, _frame: &mut eframe::Frame) {
        if self.config.side_panel_show {
            egui::SidePanel::left("LEFT_PANEL")
                .resizable(true)
                .show(ctx, |ui| {
                    ui.vertical(|ui| {
                        use utils::AmadeusMainView::*;
                        ui.style_mut().override_text_style = Some(utils::font::heading1());
                        self.render_side_panel_main_view_button(ui, "🎵 Queue", Queue);
                        self.render_side_panel_main_view_button(ui, "📚 Historic", Historic);
                        self.render_side_panel_main_view_button(ui, "🔍 Search", SearchResults);
                        ui.add_space(constants::PADDING);
                        ui.style_mut().override_text_style = Some(utils::font::body());
                        ui.separator();
                    });
                    self.playlist_store.render(ui, self.config.dark_mode);
                });
        }
        if self.playlist_store.refresh_playlists() {
            self.actions.push(actions::Action::RefreshPlaylists);
        }
    }

    fn render_central_panel(&mut self, ctx: &egui::Context, _frame: &mut eframe::Frame) {
        egui::CentralPanel::default().show(ctx, |ui| {
            use utils::AmadeusMainView::*;
            match self.config.main_panel_view {
                Queue => self.lektord_queue.render(ui, self.config.dark_mode),
                Historic => self.lektord_historic.render(ui, self.config.dark_mode),
                SearchResults => {
                    ui.style_mut().override_text_style = Some(utils::font::heading1());
                    ui.label("Database search query");
                    ui.add_space(constants::PADDING);

                    let response = ui.add(
                        egui::TextEdit::singleline(&mut self.lektord_search_query)
                            .desired_width(f32::INFINITY),
                    );
                    self.lektord_updated_query |=
                        response.lost_focus() || ui.input().key_pressed(egui::Key::Enter);
                    ui.add_space(constants::PADDING * 2.);

                    self.lektord_search_results
                        .render(ui, self.config.dark_mode);
                }
            }
        });
    }

    fn render_bottom_panel(&mut self, ctx: &egui::Context, _: &mut eframe::Frame) {
        if !(self.lektord_status.is_some() && self.lektord_current.is_some()) {
            return;
        }

        let status = self.lektord_status.as_ref().unwrap();
        let current = self.lektord_current.as_ref().unwrap();

        let progress = (status.elapsed() as f64 / status.duration() as f64) as f32;
        let text_duration = format!(
            "{:02}:{:02}",
            status.duration() / 60,
            status.duration() % 60
        );

        egui::TopBottomPanel::bottom("FOOTER")
            .max_height(constants::BOTTOM_PANEL_MAX_SIZE)
            .min_height(constants::BOTTOM_PANEL_MAX_SIZE)
            .resizable(false)
            .show(ctx, |ui| {
                // The text...
                ui.horizontal(|ui| {
                    KaraCard::new(Kara {
                        id: status.songid().unwrap() as u32,
                        source_name: current.source().to_string(),
                        song_type: current.song_type().to_string(),
                        language: current.language().to_string(),
                        category: current.category().to_string(),
                        title: current.title().to_string(),
                        song_number: current.song_number().map(|x| x as u32),
                        author: current.author().to_string(),
                        is_available: true,
                    })
                    .render_compact(ui, self.config.dark_mode);
                    ui.with_layout(egui::Layout::right_to_left(egui::Align::Min), |ui| {
                        ui.style_mut().override_text_style = Some(utils::font::heading3());
                        ui.add_space(constants::PADDING * 2.);
                        ui.label(text_duration);
                    });
                });
                // The progress bar...
                let style = ui.style().clone();
                ui.style_mut().override_text_style = Some(utils::font::small());
                ui.style_mut().visuals.override_text_color =
                    Some(constants::get_text_color(self.config.dark_mode));
                ui.add(widgets::progress_bar(self.config.dark_mode, progress));
                ui.set_style(style);
            });
    }

    fn render_top_panel(&mut self, ctx: &egui::Context, frame: &mut eframe::Frame) {
        egui::TopBottomPanel::top("MENU").show(ctx, |ui| {
            ui.add_space(constants::TOP_PANEL_PADDING * 2.);
            egui::menu::bar(ui, |ui| {
                ui.with_layout(egui::Layout::left_to_right(egui::Align::Min), |ui| {
                    ui.style_mut().override_text_style = Some(utils::font::heading1());
                    if ui.add(egui::Button::new("Amadeus").frame(false)).clicked() {
                        self.config.side_panel_show = !self.config.side_panel_show;
                    }

                    ui.add(egui::Separator::default());
                    egui::menu::menu_button(ui, "⚡", |ui| {
                        ui.style_mut().override_text_style = Some(utils::font::body());
                        ui.add(egui::Button::new("Connect lektord").wrap(false))
                            .clicked()
                            .then(|| {
                                self.actions.push(actions::Action::ConnectToLektord);
                                ui.close_menu();
                            });
                        ui.add(egui::Button::new("Disconnect lektord").wrap(false))
                            .clicked()
                            .then(|| {
                                self.actions.push(actions::Action::DisconnectFromLektord);
                                ui.close_menu();
                            });
                    })
                    .response
                    .on_hover_text("Action menu");
                    ui.add(egui::Separator::default());

                    ui.button("⚙")
                        .on_hover_text("Settings window")
                        .clicked()
                        .then(|| self.need_settings_window = true);

                    ui.button("☕")
                        .on_hover_text("About window")
                        .clicked()
                        .then(|| self.need_about_window = true);

                    ui.add(egui::Separator::default());
                    ui.button("⏪")
                        .on_hover_text("Previous")
                        .clicked()
                        .then(|| self.actions.push(actions::Action::PlaybackPrevious));
                    match self.lektord_state {
                        LektorState::Stopped | LektorState::Pause(_) => ui
                            .button("▶")
                            .on_hover_text("Play")
                            .clicked()
                            .then(|| self.actions.push(actions::Action::PlaybackPlay)),
                        LektorState::Play(_) => ui
                            .button("⏸")
                            .on_hover_text("Pause")
                            .clicked()
                            .then(|| self.actions.push(actions::Action::PlaybackPause)),
                    };
                    ui.button("⏩")
                        .on_hover_text("Next")
                        .clicked()
                        .then(|| self.actions.push(actions::Action::PlaybackNext));
                });

                ui.with_layout(egui::Layout::right_to_left(egui::Align::Min), |ui| {
                    ui.style_mut().override_text_style = Some(utils::font::body());
                    if ui.add(egui::Button::new("❌")).clicked() {
                        frame.close();
                    }

                    ui.style_mut().override_text_style = Some(utils::font::small_body());
                    let frame_duration = self
                        .begin_render_instant
                        .duration_since(self.last_render_instant)
                        .unwrap_or_default()
                        .as_millis();
                    let fps = 1000. / frame_duration as f64;
                    ui.label(format!("{fps:02.1}fps | {frame_duration:02}ms"));
                    let (drag_id, drag_space) = ui.allocate_space(ui.available_size());
                    ui.interact(drag_space, drag_id, egui::Sense::drag())
                        .dragged()
                        .then(|| frame.drag_window());
                });
            });
            ui.add_space(constants::TOP_PANEL_PADDING);
        });
    }

    fn set_visuals(&mut self, ctx: &egui::Context) {
        ctx.request_repaint();
        let mut visuals =
            either!(self.config.dark_mode => egui::Visuals::dark(); egui::Visuals::light());

        visuals.window_rounding = egui::Rounding::none();
        visuals.widgets.active.rounding = egui::Rounding::none();
        visuals.widgets.noninteractive.rounding = egui::Rounding::none();
        visuals.widgets.inactive.rounding = egui::Rounding::none();
        visuals.widgets.hovered.rounding = egui::Rounding::none();
        visuals.widgets.open.rounding = egui::Rounding::none();

        visuals.hyperlink_color = constants::get_accent_color(self.config.dark_mode);

        ctx.set_visuals(visuals);
    }

    /// Handle the events recieved from the deamons that connect to lektord.
    fn handle_deamons_events(&mut self) {
        if let Some(((_, deamon), _)) = self.deamon.as_mut() {
            while let Ok(event) = deamon.try_recv() {
                use LektorResponse::*;
                match event {
                    PlaybackStatus(_) => todo!(),
                    CurrentKara(_) => todo!(),
                    PlaylistSet(set) => {
                        self.playlist_store.clear_playlists();
                        set.into_iter()
                            .for_each(|plt| self.playlist_store.create(Playlist { name: plt }))
                    }
                    EmptyResponse(_) => todo!(),
                    KaraSet(_) => todo!(),
                }
            }
        }

        if let Some(((deamon,), _)) = self.status_deamon.as_mut() {
            while let Ok((status, current)) = deamon.try_recv() {
                // Sanitize some values, if the time/elapsed/duration is not
                // valid, we set everything to None/zero.
                self.lektord_status = then_some!(status.verify() => status);
                self.lektord_current =
                    then_some!(current.is_some() => current.unwrap().maybe_into_inner()).flatten();
            }
        }
    }

    /// Handle quits/exits of the deamons
    fn handle_deamons_exit(&mut self) {
        // Handle the deamon closing process.
        if let Some(((socket,), status_deamon)) = &self.status_deamon {
            if status_deamon.should_quit() {
                drop(socket);
                self.lektord_status = None;
                self.lektord_current = None;
                status_deamon.join();
                self.status_deamon = None;
            }
        };
        if let Some(((socket_out, socket_in), deamon)) = &self.deamon {
            if deamon.should_quit() {
                drop(socket_out);
                drop(socket_in);
                deamon.join();
                self.deamon = None;
            }
        };
    }

    /// Handle actions recieved from the user. May communicate to the deamons
    /// and send infos/commands to lektord if they are available.
    fn handle_action(&mut self) {
        use actions::Action::*;

        // Handle actions on lektor items
        for (id, act) in self.collect_fulfilled_actions() {
            match act {
                OpenPlaylist => self.playlist_store.show_playlist(id),
                ClearPlaylistContent => self.playlist_store.clear_playlist(id),
                PlayFromKara
                | DeleteKaraFromQueue
                | InsertKaraInQueue
                | AddKaraToQueue
                | AddKaraToPlaylist
                | DeleteKaraFromPlaylist(_)
                | AddPlaylistToQueue
                | InsertPlaylistToQueue => {
                    debug!("execute action {act:?} on lektor item with id {id}")
                }

                // This should not occure on items because they are global
                // actions.
                RefreshPlaylists
                | PlaybackPrevious
                | PlaybackPlay
                | PlaybackPause
                | PlaybackNext
                | ConnectToLektord
                | DisconnectFromLektord => unreachable!(),
            }
        }

        // Handle top level actions
        for act in self.actions.drain(..) {
            match act {
                ConnectToLektord => {
                    if self.status_deamon.is_none() {
                        let connexion = StatusDeamon::spawn(
                            self.config.lektord_hostname.clone(),
                            self.config.lektord_port.as_integer() as i16,
                        );
                        if let Ok(connexion) = connexion {
                            self.status_deamon = Some(connexion);
                        }
                    }
                    if self.deamon.is_none() {
                        let connexion = CommandDeamon::spawn(
                            self.config.lektord_hostname.clone(),
                            self.config.lektord_port.as_integer() as i16,
                        );
                        if let Ok(connexion) = connexion {
                            self.deamon = Some(connexion);
                        }
                    }
                }
                DisconnectFromLektord => {
                    debug!("asked to close the deamons");
                    if let Some((_, status_deamon)) = &self.status_deamon {
                        status_deamon.quit();
                    }
                    if let Some((_, deamon)) = &self.deamon {
                        deamon.quit();
                    }
                    break;
                }

                PlaybackPrevious | PlaybackPlay | PlaybackPause | PlaybackNext => {
                    debug!("Execute action {act:?} on lektor")
                }

                RefreshPlaylists => {
                    if let Some(((sender, _), _)) = &self.deamon {
                        handle_sender_error(sender.send(LektorQuery::ListAllPlaylists))
                    }
                }

                _ => unreachable!(),
            }
        }
    }

    fn apply_settings(&mut self) {
        self.lektord_queue
            .with_max_content(self.config.ui_max_queue_items.as_integer());
        self.lektord_historic
            .with_max_content(self.config.ui_max_history_items.as_integer());
        self.lektord_search_results
            .with_max_content(self.config.ui_max_search_items.as_integer());
    }
}

impl App for Amadeus<'_> {
    fn update(&mut self, ctx: &egui::Context, frame: &mut eframe::Frame) {
        self.begin_render_instant = time::SystemTime::now();
        ctx.request_repaint();
        self.set_visuals(ctx);

        self.render_top_panel(ctx, frame);
        self.render_side_panel(ctx, frame);
        self.render_central_panel(ctx, frame);
        self.render_bottom_panel(ctx, frame);

        self.playlist_store
            .render_playlist_contents(ctx, self.config.dark_mode);
        widgets::render_about_window(
            &mut self.need_about_window,
            self.config.dark_mode,
            ctx,
            self.amadeus_logo_texture.as_ref().unwrap(),
        );
        self.config.render_settings_window(
            ctx,
            &mut self.need_settings_window,
            &mut self.has_config_changed,
        );

        self.last_render_instant = self.begin_render_instant;

        self.handle_deamons_events();
        self.handle_deamons_exit();
        self.handle_action();

        if self.has_config_changed {
            self.apply_settings();
        }
    }

    fn save(&mut self, storage: &mut dyn eframe::Storage) {
        if self.has_config_changed {
            eframe::set_value(storage, "amadeus-rs-config", &self.config);
            self.has_config_changed = false;
        }
    }

    fn persist_egui_memory(&self) -> bool {
        true
    }

    fn persist_native_window(&self) -> bool {
        true
    }

    fn auto_save_interval(&self) -> time::Duration {
        time::Duration::from_secs(1)
    }
}
