//! Common types for lektord.

use std::hash::{Hash, Hasher};

/// Lektor Types are the Kara and the Playlist types. You should not implement
/// this type yourself.
pub trait LektorType<'a>:
    Clone + serde::Serialize + serde::Deserialize<'a> + Hash + private::Sealed
{
    fn unique_id(&self) -> u64;
}

#[derive(Clone, Eq, serde::Serialize, serde::Deserialize)]
pub struct Kara {
    pub id: u32,
    pub source_name: String,

    #[serde(rename = "type")]
    pub song_type: String,

    pub language: String,
    pub category: String,
    pub title: String,
    pub song_number: Option<u32>,
    pub author: String,
    pub is_available: bool,
}

#[derive(Clone, Eq, serde::Serialize, serde::Deserialize, PartialEq, Hash)]
pub struct Playlist {
    pub name: String,
}

impl PartialEq for Kara {
    fn eq(&self, other: &Self) -> bool {
        self.id == other.id
    }
}

impl LektorType<'_> for Kara {
    fn unique_id(&self) -> u64 {
        private::unique_id(self)
    }
}

impl LektorType<'_> for Playlist {
    fn unique_id(&self) -> u64 {
        private::unique_id(self)
    }
}

#[derive(serde::Serialize, serde::Deserialize, Clone, Debug, PartialEq, Eq, Copy)]
pub enum LektorState {
    Stopped,
    Play(usize),
    Pause(usize),
}

impl Default for LektorState {
    fn default() -> Self {
        Self::Stopped
    }
}

/// We seal the implementation only for supported types.
mod private {
    use super::{Kara, Playlist};
    use std::{
        collections::hash_map::DefaultHasher,
        hash::{Hash, Hasher},
    };

    pub trait Sealed {}

    impl Sealed for Kara {}
    impl Sealed for Playlist {}

    pub(crate) fn unique_id<T: Hash + Sealed>(object: &T) -> u64 {
        let mut s = DefaultHasher::new();
        object.hash(&mut s);
        s.finish()
    }
}

/// Sealed implementation of Hash for Kara.
impl Hash for Kara {
    fn hash<H: Hasher>(&self, state: &mut H) {
        ((self.id as usize) * 2).hash(state);
    }
}
